import os
import shutil
import re
import glob
import bisect
from datetime import date
from enum import Enum
from kabaret import flow
from kabaret.flow_entities.entities import Entity
from libreflow.baseflow.film import Film as BaseFilm
from libreflow.utils.os import zip_folder

from .shot import Shots


class ShotState(Enum):

    MISSING_FILE                = 0
    FIRST_DELIVERY              = 1  # First delivery
    NEW_TAKE_PACK_SELECTED      = 5  # New take + resend pack
    NEW_TAKE_PACK_UNSELECTED    = 6
    RESEND_TAKE_PACK_SELECTED   = 11 # Resend last take + resend pack
    RESEND_TAKE_PACK_UNSELECTED = 12
    ALREADY_DELIVERED           = 7
    ALREADY_RECEIVED            = 8


class ShotToSend(flow.SessionObject):

    sequence_name  = flow.Param()
    shot_name      = flow.Param()
    last_take      = flow.Param()
    target_take    = flow.Computed()

    # state          = flow.Param()
    selected       = flow.Param().ui(editor='bool')
    new_take       = flow.Param().ui(editor='bool')
    send_pack      = flow.Param().ui(editor='bool')
    # message        = flow.Param()

    _action        = flow.Parent(2)

    def get_init_state(self):
        return self._action.get_init_state(self.name())
    
    def ensure_take(self, kitsu_task_name):
        return self._action.ensure_take(
            self.name(),
            'working_file_plas',
            kitsu_task_name,
            take_name=self.target_take.get(),
            replace_take=not self.new_take.get()
        )
    
    def create_package(self, delivery_dir=None):
        self._action.create_shot_package(
            self.name(),
            self._action.task.get(),
            self.target_take.get(),
            self.send_pack.get(),
            delivery_dir,
        )
    
    def update_kitsu_status(self, kitsu_task_name, kitsu_status_name, revision_name):
        self._action.update_kitsu_status(
            self.name(),
            kitsu_task_name,
            kitsu_status_name,
            self.target_take.get(),
            revision_name
        )

    def compute_child_value(self, child_value):
        if child_value is self.target_take:
            last_take = self.last_take.get()
            if last_take is None:
                index = 1
            else:
                m = re.match(r't(\d+)', last_take)
                if m is not None:
                    index = int(m.group(1))
                    if self.new_take.get():
                        index += 1
                else:
                    index = 1
            self.target_take.set('t'+str(index))


class ShotsToSend(flow.DynamicMap):

    _action = flow.Parent()

    @classmethod
    def mapped_type(cls):
        return ShotToSend

    def __init__(self, parent, name):
        super(ShotsToSend, self).__init__(parent, name)
        self._shot_names = None
        self._shot_cache = None
    
    def mapped_names(self, page_num=0, page_size=None):
        kitsu_status = self._action.to_send_kitsu_status.get()
        
        if self._shot_names is None:
            self._mng.children.clear()
            self._shot_names = []
            self._shot_cache = {}
            
            # Get shots ready to be sent on Kitsu
            kitsu = self.root().project().kitsu_api()
            validated_shots = kitsu.get_shots({
                self._action.task.get(): [self._action.to_send_kitsu_status.get()]
            })
            for sq, sh in validated_shots:
                n = f'{sq}{sh}'
                
                bisect.insort(self._shot_names, n)
                self._shot_cache[n] = dict(
                    sequence_name=sq,
                    shot_name=sh,
                    last_take_name=self._action.get_last_take(n)
                )
        
        return self._shot_names
    
    def touch(self):
        self._shot_names = None
        super(ShotsToSend, self).touch()
    
    def _configure_child(self, item):
        self.mapped_names()
        data = self._shot_cache[item.name()]
        item.sequence_name.set(data['sequence_name'])
        item.shot_name.set(data['shot_name'])
        item.last_take.set(data['last_take_name'])


class KitsuTaskNames(flow.values.SessionValue):

    DEFAULT_EDITOR = 'choice'

    def choices(self):
        return ['L&S KEY FRAME', 'L&S ANIMATION']


class SendForValidation(flow.Action):

    FILES_BY_TASK = {
        'L&S KEY FRAME': ('takes_fix.plas', 'takes_fix', None),
        'L&S ANIMATION': ('takes_ani.plas', 'takes_ani', 'ani'),
    }

    to_send_kitsu_status = flow.Param('H TO SEND').ui(hidden=True)
    task                 = flow.SessionParam('L&S KEY FRAME', KitsuTaskNames).watched()
    task_status          = flow.Param('I Waiting For Approval').ui(hidden=True)
    target_site          = flow.Param('test2').ui(hidden=True)
    
    shots = flow.Child(ShotsToSend)

    _film = flow.Parent()

    def needs_dialog(self):
        self.message.set('<h2>Send shots for validation</h2>')
        self.task.revert_to_default()
        return True
    
    def run(self, button):
        task_name = self.task.get()
        task_status_name = self.task_status.get()
        delivery_dir = self.get_delivery_dir()

        delivered_shots = []
        take_prefix = self.FILES_BY_TASK[self.task.get()][2] or ''
        if take_prefix:
            take_prefix += '_'

        for s in self.shots.mapped_items():
            if s.selected.get():
                take_name, source_revision_name = s.ensure_take(task_name)
                s.create_package(delivery_dir)
                s.update_kitsu_status(task_name, task_status_name, source_revision_name)
                delivered_shots.append((s.name(), f'{take_prefix}{take_name}'))
        
        self.create_report(delivery_dir, delivered_shots)

    def ensure_take(self, shot_name, working_file_name, kitsu_task_name, take_name=None, replace_take=False):
        lighting_files = self._film.shots[shot_name].tasks['lighting'].files
        file_name, file_display_name, suffix = self.FILES_BY_TASK[kitsu_task_name]
        suffix = suffix and suffix+'_' or ''
        name, ext = file_name.split('.')

        if not lighting_files.has_file(name, ext):
            take_file = lighting_files.add_file(
                name, ext,
                display_name=file_display_name,
                tracked=True,
                default_path_format='{sequence}_{shot}/{sequence}_{shot}_'+suffix+'{revision}'
            )
        else:
            take_file = lighting_files[f'{name}_{ext}']
        
        source_revision = lighting_files[working_file_name].get_head_revision()
        source_path = source_revision.get_path()

        if replace_take:
            if take_name is None:
                take = take_file.get_head_revision()
            else:
                take = take_file.get_revision(take_name)
            
            take.comment.set(f'Created from working file {source_revision.name()}')
            os.remove(take.get_path())
        else:
            if take_name is None:
                take_index = len(take_file.get_revisions().mapped_names()) + 1
                take_name = 't'+str(take_index)
            
            take = take_file.add_revision(
                name=take_name,
                comment=f'Created from working file {source_revision.name()}'
            )

        take_path = take.get_path()
        os.makedirs(os.path.dirname(take_path), exist_ok=True)
        shutil.copy2(source_path, take_path)

        lighting_files.touch()

        return take.name(), source_revision.name()
    
    def create_shot_package(self, shot_name, kitsu_task_name, take_name, include_pack=False, delivery_dir=None):
        lighting_files = self._film.shots[shot_name].tasks['lighting'].files
        delivery_files = self._film.shots[shot_name].tasks['delivery'].files
        today = date.today().strftime('%y%m%d')
        package_name = f'package_{today}'

        # Create package folder
        if not delivery_files.has_folder(package_name):
            package = delivery_files.add_folder(
                package_name,
                tracked=True,
                default_path_format='{sequence}_{shot}/package/'+today+'/{revision}'
            )
        else:
            package = delivery_files[package_name]
        
        pkg_rev = package.add_revision()
        pkg_path = pkg_rev.get_path()
        os.makedirs(pkg_path)

        # Copy working file in package folder
        file_name, file_display_name, suffix = self.FILES_BY_TASK[kitsu_task_name]
        name, ext = file_name.split('.')
        suffix = suffix and suffix+'_' or ''

        take_file = lighting_files[f'{name}_{ext}']
        take_path = take_file.get_revision(take_name).get_path()

        self.root().session().log_info(
            f'Copy take {take_path} into package {pkg_path}'
        )
        shutil.copy2(take_path, pkg_path)

        # If asked, copy pack as well
        if include_pack:
            pack_path = lighting_files['pack'].get_head_revision().get_path()
            self.root().session().log_info(
                f'Copy pack {pack_path} into package {pkg_path}'
            )
            shutil.copytree(pack_path, os.path.join(pkg_path, 'pack'))
        
        # Upload package
        current_site = self.root().project().get_current_site()
        upload_job = current_site.get_queue().submit_job(
            emitter_oid=pkg_rev.oid(),
            user=self.root().project().get_user_name(),
            studio=current_site.name(),
            job_type='Upload',
            init_status='WAITING'
        )

        sync_mgr = self.root().project().get_sync_manager()
        sync_mgr.process(upload_job)
        
        # Zip package in given delivery folder
        if delivery_dir is not None:
            pkg_name = f'{shot_name}_rouge_{suffix}{take_name}.zip'     # <shot_name>_rouge_<take_name>.zip
            delivery_pkg_path = os.path.join(
                delivery_dir, pkg_name
            )
            zip_folder(pkg_path, delivery_pkg_path)
        else:
            self.root().session().log_warning(
                f'{shot_name} {take_name} :: delivery folder {delivery_dir} not found'
            )
        
        msg = f'{shot_name} {take_name} :: package uploaded'
        
        self.root().session().log_info(msg)
    
    def update_kitsu_status(self, shot_name, kitsu_task_name, kitsu_status_name, take_name, revision_name):
        user_name = self.root().project().get_user_name()
        take_file_name = self.FILES_BY_TASK[kitsu_task_name][1]
        shot = self._film.shots[shot_name]
        task = shot.tasks['lighting']

        shot.set_task_status(
            kitsu_task_name,
            kitsu_status_name,
            comment=(
                f"**{user_name}** has submitted take **{take_name}** for validation.\n"
                f"- Created from working file **{revision_name}**\n\n"
                f"*{task.oid()}*"
            )
        )
    
    def get_init_state(self, shot_name):
        state = ShotState.MISSING_FILE
        if (
            self._film.shots[shot_name].tasks.has_mapped_name('lighting')
            and self._film.shots[shot_name].tasks['lighting'].files.has_file('working_file', 'plas')
            and self._film.shots[shot_name].tasks['lighting'].files['working_file_plas'].get_head_revision() is not None
            and self._film.shots[shot_name].tasks['lighting'].files['working_file_plas'].get_head_revision().get_sync_status() == 'Available'
        ):
            files = self._film.shots[shot_name].tasks['lighting'].files

            if (
                self.task.get() == 'L&S KEY FRAME'
                and (
                    not files.has_file('takes_fix', 'plas')
                    or not files['takes_fix_plas'].has_revision('t1')
                )
            ):
                state = ShotState.FIRST_DELIVERY
            else:
                today = date.today().strftime('%y%m%d')
                package_name = f'package_{today}'
                if (
                    self._film.shots[shot_name].tasks['delivery'].files.has_folder(package_name)
                    and self._film.shots[shot_name].tasks['delivery'].files[package_name].get_head_revision() is not None
                    and self._film.shots[shot_name].tasks['delivery'].files[package_name].get_head_revision().get_sync_status() == 'Available'
                ):
                    last_package = self._film.shots[shot_name].tasks['delivery'].files[package_name].get_head_revision()
                    
                    if last_package.get_sync_status(site_name=self.target_site.get()) != 'Available':
                        state = ShotState.ALREADY_DELIVERED
                    else:
                        state = ShotState.ALREADY_RECEIVED
                else:
                    state = ShotState.NEW_TAKE_PACK_UNSELECTED
        
        return state
    
    def get_last_take(self, shot_name):
        last_take = None
        file_name, file_display_name, suffix = self.FILES_BY_TASK[self.task.get()]
        name, ext = file_name.split('.')

        if (
            self._film.shots[shot_name].tasks.has_mapped_name('lighting')
            and self._film.shots[shot_name].tasks['lighting'].files.has_file(name, ext)
        ):
            last_take = self._film.shots[shot_name].tasks['lighting'].files[f'{name}_{ext}'].get_head_revision()
            if last_take is not None:
                last_take = last_take.name()
        
        return last_take

    def get_delivery_dir(self):
        delivery_dir = self.root().project().get_current_site().delivery_dir.get()

        if delivery_dir is not None:
            today = date.today().strftime('%y%m%d')
            delivery_dir = os.path.join(delivery_dir, today)

            if os.path.isdir(delivery_dir):
                i = 2
                while os.path.isdir(f'{delivery_dir}-{i}'):
                    i += 1
                delivery_dir = f'{delivery_dir}-{i}'
            
            os.makedirs(delivery_dir, exist_ok=True)
        
        return delivery_dir
    
    def create_report(self, delivery_dir, delivered_shots):
        if not delivered_shots:
            return
        
        with open(os.path.join(delivery_dir, 'delivery_report.txt'), 'w+') as report_file:
            for shot_name, take_name in delivered_shots:
                report_file.write(f'{shot_name} {take_name}\n')
    
    def child_value_changed(self, child_value):
        if child_value is self.task:
            self.shots.touch()
    
    def _fill_ui(self, ui):
        ui['custom_page'] = 'libreflow.rouge.ui.package.SendShotsForValidationWidget'


class AddPresetAction(flow.Action):

    ICON = ('icons.gui', 'plus-sign-in-a-black-circle')

    # Maybe set to SessionParam
    preset_name          = flow.Param('').ui(label='Name')
    task_name            = flow.Param('')
    path_template        = flow.Param('')
    target_name          = flow.Param('')
    with flow.group('Optional'):
        optional_init    = flow.Param().ui(editor='bool').ui(label='When init')
        optional_new_rev = flow.Param().ui(editor='bool').ui(label='When new revision')
    no_track             = flow.BoolParam()

    _map = flow.Parent()

    def get_buttons(self):
        return ('Add', 'Cancel')
    
    def run(self, button):
        if button == 'Cancel':
            return
        
        p = self._map.add(self.preset_name.get())
        p.task_name.set(self.task_name.get())
        p.path_template.set(self.path_template.get())
        p.target_name.set(self.target_name.get())
        p.optional_init.set(self.optional_init.get())
        p.optional_new_rev.set(self.optional_new_rev.get())
        p.no_track.set(self.no_track.get())
        self._map.touch()
        return self.get_result(close=True)


class FileSourcePreset(flow.Object):

    task_name            = flow.Param('')
    path_template        = flow.Param() # e.g., "^rgr(\D*\d){2,}_(\d{3})(_*){1}.plas"
    target_name          = flow.Param('')
    with flow.group('Optional'):
        optional_init    = flow.BoolParam().ui(label='When init')
        optional_new_rev = flow.BoolParam().ui(label='When new revision')
    no_track             = flow.BoolParam()


class FileSourcePresets(flow.Map):

    add_preset = flow.Child(AddPresetAction)

    @classmethod
    def mapped_type(cls):
        return FileSourcePreset


class FilesItem(flow.Object):

    file_name            = flow.Param('')
    file_format          = flow.Param('')
    file_comment         = flow.Param('')
    file_path            = flow.Param('')
    file_target          = flow.Param('')
    rev_version          = flow.Param('v001')
    with flow.group('Optional'):
        optional_init    = flow.BoolParam().ui(label='When init')
        optional_new_rev = flow.BoolParam().ui(label='When new revision')


class AddFileTest(flow.Action):

    file_name           = flow.Param('')
    file_comment        = flow.Param('')
    file_path           = flow.Param('')
    file_target         = flow.Param('')
    rev_version         = flow.Param('v001')
    
    files_map           = flow.Parent()

    def needs_dialog(self):
        return True
    
    def get_buttons(self):
        return ['Add', 'Cancel']

    def run(self, button):
        if button == 'Cancel':
            return
        
        display_name = self.file_name.get().replace(".", "_")
        item = self.files_map.add(display_name)
        item.file_name.set(self.file_name.get())
        item.file_comment.set(self.file_comment.get())
        item.file_path.set(self.file_path.get())
        item.file_target.set(self.file_target.get())
        item.rev_version.set(self.rev_version.get())

        self.files_map.touch()
        return self.get_result(close=True)


class FilesMap(flow.Map):

    add_file            = flow.Child(AddFileTest)

    @classmethod
    def mapped_type(cls):
        return FilesItem

    def columns(self):
        return ["Name", "Target", "Revision"]

    def _configure_child(self, child):
        child.file_comment.set('')

    def _fill_row_cells(self, row, item):
        row["Name"] = item.file_name.get()
        row["Target"] = item.file_target.get()
        row["Revision"] = item.rev_version.get()


class ShotItem(flow.Object):

    shot_name           = flow.Param('')
    new_shot            = flow.BoolParam(False)
    shot_status         = flow.Param('')
    files_map           = flow.Child(FilesMap)


class AddShotTest(flow.Action):

    shot_name           = flow.Param('')
    new_shot            = flow.BoolParam(False)
    shot_status         = flow.Param('')
    
    shots_map           = flow.Parent()

    def needs_dialog(self):
        return True
    
    def get_buttons(self):
        return ['Add', 'Cancel']

    def run(self, button):
        if button == 'Cancel':
            return
        
        item = self.shots_map.add(self.shot_name.get())
        item.shot_name.set(self.shot_name.get())
        item.new_shot.set(self.new_shot.get())
        item.shot_status.set(self.shot_status.get())

        self.shots_map.touch()
        return self.get_result(close=True)


class ShotsMap(flow.Map):

    add_shot            = flow.Child(AddShotTest)

    @classmethod
    def mapped_type(cls):
        return ShotItem

    def columns(self):
        return ["Name", "New Shot"]

    def reset_all(self):
        for item in self.mapped_items():
            item.files_map.clear()
        self.clear()

    def _configure_child(self, child):
        child.shot_name.set(child.name())

    def _fill_row_cells(self, row, item):
        row["Name"] = item.shot_name.get()
        row["New Shot"] = item.new_shot.get()


class ImportFiles(flow.Action):

    ICON = ('icons.gui', 'file')

    shots_map           = flow.Child(ShotsMap)

    # Settings (les passer en SessionParam)
    file_source_presets = flow.Child(FileSourcePresets)
    shot_name_pattern   = flow.Param('^rgr(\D*\d){2,}_(\d{3})')
    can_edit_target     = flow.BoolParam(False)
    change_kitsu_status = flow.BoolParam(False)
    target_kitsu_task   = flow.DictParam(dict(name='PREPARATION PALAS', status='DONE'))

    _film               = flow.Parent()

    def clear_list(self):
        self.shots_map.reset_all()

    def get_user_status(self):
        name = self.root().project().get_user_name()
        return self.root().project().get_user(name).status.get()

    def get_shot(self, name):
        kitsu_shot_names = self.get_kitsu_shot_names(name)

        if kitsu_shot_names is None:
            return False
        
        kitsu_config = self.root().project().kitsu_api()
        sequence_name = list(kitsu_shot_names.keys())[0]
        
        for shot_name in kitsu_shot_names[sequence_name]:
            if kitsu_config.get_shot_data(shot_name, sequence_name) is None:
                return False

        # Check object
        if self._film.shots.has_mapped_name(name) == False:
            return 'Create'

        return self._film.shots[name]

    def check_files(self, paths):
        incorrect = {
            'files': [],
            'shots': []
        }
        valid = []

        for path in paths:
            rev = None
            status = None
            shot_files = []
            if os.path.isfile(path):
                #region Check if plas file
                file_fullname = path.split('/')[-1]
                m = None
                preset = None
                for p in self.file_source_presets.mapped_items():
                    regex = p.path_template.get()
                    m = re.search(regex, file_fullname, re.IGNORECASE)
                    if m:
                        preset = p
                        break
                # unknown file
                if m is None:
                    incorrect['files'].append({
                        'name': '/'.join(path.split('/')[-2:]),
                        'error': 'file'
                    })
                    continue
                #endregion

                #region Call get_shot
                shot_name = file_fullname.split('.')[0]
                if shot_name.endswith('_') and shot_name[-3] == 't':
                    shot_name = '_'.join(shot_name.split('_')[:-2])
                elif shot_name.endswith('_'):
                    shot_name = shot_name[:-1]
                shot = self.get_shot(shot_name)
                if shot == False:
                    incorrect['files'].append({
                        'name': '/'.join(path.split('/')[-2:]),
                        'error': 'kitsu'
                    })
                    continue
                if shot == 'Create':
                    if p.optional_init.get() == False:
                        incorrect['files'].append({
                            'name': '/'.join(path.split('/')[-2:]),
                            'error': 'init'
                        })
                        continue
                #endregion

                #region Check revision
                # For generic version
                # file_name = file_fullname.replace('.', '_')
                # file_obj = shot.tasks[preset.task_name.get()].files.has_mapped_name(file_name)
                file_name = file_fullname.split('.')[0]
                target_name = preset.target_name.get()
                if target_name == '':
                    target_name = file_fullname
                if '.' in target_name:
                    target_name = target_name.replace('.', '_')

                #region rtk status
                if shot != 'Create':
                    if shot.tasks[preset.task_name.get()].files.has_mapped_name(target_name) == True:
                        status = 'rtk'
                        rev = shot.tasks[preset.task_name.get()].files[target_name].get_head_revision()
                        if rev and file_name.endswith('_') == False or (file_name.endswith('_') == False and file_name[-3] != 't'):
                            if preset.optional_new_rev.get() == False:
                                incorrect['files'].append({
                                    'name': '/'.join(path.split('/')[-2:]),
                                    'error': 'rtk'
                                })
                                continue
                #endregion
                
                #region init status
                    else:
                        status = 'init'
                        if preset.optional_init.get() == False:
                            incorrect['files'].append({
                                'name': '/'.join(path.split('/')[-2:]),
                                'error': 'init'
                            })
                            continue
                
                if shot == 'Create':
                    status = 'init'
                #endregion

                #endregion
                    
                #region Add in valid list
                # Add in correct shot object
                find_shot = False
                for s in valid:
                    if s['shot'] == shot_name:
                        find_shot = True
                        s['files'].append({
                            'path': path,
                            'preset': preset,
                            'name': file_fullname
                        })
                        break
                if find_shot == False:
                    valid.append({
                        'shot': shot_name,
                        'obj': shot,
                        'status': status,
                        'files': [{
                            'path': path,
                            'preset': preset,
                            'name': file_fullname
                        }]
                    })
                #endregion
            if os.path.isdir(path):
                #region Check dir name
                dir_name = path.split('/')[-1]
                regex = self.shot_name_pattern.get()
                m = re.search(regex, dir_name, re.IGNORECASE)
                if m is None:
                    if dir_name == 'pack':
                        incorrect['files'].append({
                            'name': '/'.join(path.split('/')[-2:]),
                            'error': 'pack_only'
                        })
                    else:
                        incorrect['files'].append({
                            'name': '/'.join(path.split('/')[-2:]),
                            'error': 'dir_name'
                        })
                    continue
                #endregion

                #region Call get_shot
                shot = self.get_shot(dir_name)
                if shot == False:
                    find_entry = False
                    for s in incorrect['shots']:
                        if s['shot'] == dir_name:
                            find_entry = True
                            s['error'] = 'kitsu'
                            break
                    if find_entry == False:
                        incorrect['shots'].append({
                            'shot': dir_name,
                            'files': [],
                            'error': 'kitsu'
                        })
                    continue
                #endregion

                #region Check if dir is empty
                if os.listdir(path) == []:
                    find_entry = False
                    for s in incorrect['shots']:
                        if s['shot'] == dir_name:
                            find_entry = True
                            s['error'] = 'dir_empty'
                            break
                    if find_entry == False:
                        incorrect['shots'].append({
                            'shot': dir_name,
                            'files': [],
                            'error': 'dir_empty'
                        })
                    continue
                #endregion
                
                #region listdir
                for file_fullname in os.listdir(path):
                    m = None
                    preset = None
                    #region Check targets
                    for p in self.file_source_presets.mapped_items():
                        regex = p.path_template.get()
                        m = re.search(regex, file_fullname, re.IGNORECASE)
                        if m:
                            preset = p
                            break
                    file_path = os.path.join(path, file_fullname)
                    file_name = file_fullname.split('.')[0]
                    # unknown file
                    if m is None:
                        find_entry = False
                        for s in incorrect['shots']:
                            if s['shot'] == dir_name:
                                find_entry = True
                                s['files'].append({
                                    'name': file_fullname,
                                    'error': 'file'
                                })
                                break
                        if find_entry == False:
                            incorrect['shots'].append({
                                'shot': dir_name,
                                'files': [{
                                    'name': file_fullname,
                                    'error': 'file'
                                }],
                                'error': ''
                            })
                        continue
                    #endregion
                    
                    #region Check for pack
                    if 'pack' in file_fullname:
                        if shot != 'Create':
                            if shot.tasks[preset.task_name.get()].files.has_mapped_name(file_fullname) == True:
                                if preset.no_track.get() == True:
                                    find_entry = False
                                    for s in incorrect['shots']:
                                        if s['shot'] == dir_name:
                                            find_entry = True
                                            s['files'].append({
                                                'name': file_fullname,
                                                'error': 'no_track'
                                            })
                                            break
                                    if find_entry == False:
                                        incorrect['shots'].append({
                                            'shot': dir_name,
                                            'files': [{
                                                'name': file_fullname,
                                                'error': 'no_track'
                                            }],
                                            'error': ''
                                        })
                                    continue
                                else:
                                    find_entry = False
                                    for s in incorrect['shots']:
                                        if s['shot'] == dir_name:
                                            find_entry = True
                                            s['files'].append({
                                                'name': file_fullname,
                                                'error': 'overwrite'
                                            })
                                            break
                                    if find_entry == False:
                                        incorrect['shots'].append({
                                            'shot': dir_name,
                                            'files': [{
                                                'name': file_fullname,
                                                'error': 'overwrite'
                                            }],
                                            'error': ''
                                        })

                    #endregion

                    #region Checks for plas file
                    if '.plas' in file_fullname:
                        #region if plas name == dir name
                        if file_name.endswith('_') and file_name[-3] == 't':
                            file_name = '_'.join(file_name.split('_')[:-2])
                        elif file_name.endswith('_'):
                            file_name = file_name[:-1]
                        if file_name != dir_name:
                            find_entry = False
                            for s in incorrect['shots']:
                                if s['shot'] == dir_name:
                                    find_entry = True
                                    s['files'].append({
                                        'name': file_fullname,
                                        'error': 'not_match'
                                    })
                                    break
                            if find_entry == False:
                                incorrect['shots'].append({
                                    'shot': dir_name,
                                    'files': [{
                                        'name': file_fullname,
                                        'error': 'not_match'
                                    }],
                                    'error': ''
                                })
                            continue
                        #endregion

                        #region Check revision
                        file_name = file_fullname.split('.')[0]
                        target_name = preset.target_name.get()
                        if target_name == '':
                            target_name = file_fullname
                        if '.' in target_name:
                            target_name = target_name.replace('.', '_')
                        #region rtk status
                        if shot != 'Create':
                            if shot.tasks[preset.task_name.get()].files.has_mapped_name(target_name) == True:
                                status = 'rtk'
                                rev = shot.tasks[preset.task_name.get()].files[target_name].get_head_revision()
                                if rev and file_name.endswith('_') == False or (file_name.endswith('_') == False and file_name[-3] != 't'):
                                    if preset.optional_new_rev.get() == False:
                                        find_entry = False
                                        for s in incorrect['shots']:
                                            if s['shot'] == dir_name:
                                                find_entry = True
                                                s['files'].append({
                                                    'name': file_fullname,
                                                    'error': 'rtk'
                                                })
                                                break
                                        if find_entry == False:
                                            incorrect['shots'].append({
                                                'shot': dir_name,
                                                'files': [{
                                                    'name': file_fullname,
                                                    'error': 'rtk'
                                                }],
                                                'error': ''
                                            })
                                        continue
                        #endregion

                        #region init status
                            else:
                                status = 'init'
                                if file_name.endswith('_') or (file_name.endswith('_') and file_name[-3] == 't'):
                                    if preset.optional_init.get() == False:
                                        find_entry = False
                                        for s in incorrect['shots']:
                                            if s['shot'] == dir_name:
                                                find_entry = True
                                                s['files'].append({
                                                    'name': file_fullname,
                                                    'error': 'init'
                                                })
                                                break
                                        if find_entry == False:
                                            incorrect['shots'].append({
                                                'shot': dir_name,
                                                'files': [{
                                                    'name': file_fullname,
                                                    'error': 'init'
                                                }],
                                                'error': ''
                                            })
                                        continue
                        
                        if shot == 'Create':
                            status = 'init'
                            if file_name.endswith('_') or (file_name.endswith('_') and file_name[-3] == 't'):
                                if preset.optional_init.get() == False:
                                    find_entry = False
                                    for s in incorrect['shots']:
                                        if s['shot'] == dir_name:
                                            find_entry = True
                                            s['files'].append({
                                                'name': file_fullname,
                                                'error': 'init'
                                            })
                                            break
                                    if find_entry == False:
                                        incorrect['shots'].append({
                                            'shot': dir_name,
                                            'files': [{
                                                'name': file_fullname,
                                                'error': 'init'
                                            }],
                                            'error': ''
                                        })
                                    continue

                        #endregion

                        #endregion
                    
                    #endregion
                    
                    shot_files.append({
                        'path': file_path,
                        'preset': preset,
                        'name': file_fullname
                    })
                #endregion
                
                #region no valid files
                if shot_files == []:
                    for s in incorrect['shots']:
                        if s['shot'] == dir_name:
                            s['error'] = 'no_valid_files'
                    continue
                #endregion

                #region all files is here ?
                # voir plus tard pour remettre au propre les paramètres

                #region init status
                all_files = True
                if shot == 'Create' or shot.tasks['lighting'].files.has_mapped_name('working_file_plas') == False:
                    for p in self.file_source_presets.mapped_items():
                        find_file = False
                        for f in shot_files:
                            if f['preset'] == p:
                                f = f['path']
                                find_file = True
                                break
                        if find_file == False:
                            if p.optional_init.get() == True:
                                continue
                            else:
                                all_files = False
                                break
                        continue
                #endregion
                
                #region rtk status
                else:
                    for p in self.file_source_presets.mapped_items():
                        find_file = False
                        for f in shot_files:
                            if f['preset'] == p:
                                find_file = True
                                break
                        if find_file == False:
                            if p.optional_new_rev.get() == True:
                                continue
                            else:
                                all_files = False
                                break
                        continue
                #endregion

                #region add to valid list
                if all_files == True:
                    print('SHOT OK!')

                    valid.append({
                        'shot': dir_name,
                        'obj': shot,
                        'status': status,
                        'files': shot_files
                    })
                else:
                    if shot_files != []:
                        print('SHOT NOT OKAY!')
                        for s in incorrect['shots']:
                            if s['shot'] == dir_name:
                                s['error'] = 'not_all_files'

                #endregion

                #endregion
        
        # Correct files
        # Create objects in maps
        for i in valid:
            # Create shot if needed
            if self.shots_map.has_mapped_name(i['shot']) == True:
                shot = self.shots_map[i['shot']]
            else:
                shot = self.shots_map.add(i['shot'])

            if i['obj'] == 'Create':
                shot.new_shot.set(True)
            
            shot.shot_status.set(i['status'])
            
            #Create files in shot
            for f in i['files']:
                name = f['name'].replace('.', '_')
                # Ignore if already add
                if shot.files_map.has_mapped_name(name) == True:
                    continue
                # Folder has a base and retake file
                if shot.files_map.has_mapped_name(i['shot'] + '_plas') == True:
                    name, ext = f['name'].rsplit('.', 1)
                    if name.endswith('_') or (name.endswith('_') and name[-3] == 't'):
                        continue

                f_item = shot.files_map.add(name)
                f_item.file_name.set(f['name'])

                if i['status'] == 'rtk':
                    if '.' in f['name']:
                        f_item.file_comment.set('Created from ' + f['name'])

                root, ext = os.path.splitext(f['path'])
                if ext:
                    f_item.file_format.set(ext)
                
                f_item.file_path.set(f['path'])
                
                target_name = f['preset'].target_name.get()
                if target_name == '':
                    target_name = name
                if f['preset'].task_name.get() != '':
                    f_item.file_target.set(f['preset'].task_name.get() + '/' + target_name)
                else:
                    f_item.file_target.set(target_name)
                
                if i['obj'] == 'Create':
                    f_item.rev_version.set('v001')
                else:
                    if '.' in target_name:
                        target_name = target_name.replace('.', '_')
                    if i['obj'].tasks[f['preset'].task_name.get()].files.has_mapped_name(target_name):
                        rev = i['obj'].tasks[f['preset'].task_name.get()].files[target_name].get_head_revision()
                        ver = int(rev.name().replace('v', ''))
                        ver = 'v' + f'{(ver + 1):03}'
                        f_item.rev_version.set(ver)
                
                if f['preset'].optional_init.get() == True:
                    f_item.optional_init.set(True)
                if f['preset'].optional_new_rev.get() == True:
                    f_item.optional_new_rev.set(True)

        # print('Incorrect files=> ', incorrect)
        # print('Valid shots=> ', valid)
        return incorrect
    
    def get_kitsu_shot_names(self, name):
        m = re.match('^(rgr\d\d+_)(\d{3})(_\d{3})?(_\d{3})?(_\d{3})?', name)

        if m is None:
            return None
        
        sequence_name = m.group(1)
        shot_names = [m.group(2)]
        for i in range(3, len(m.groups()) + 1):
            if m.group(i) is None:
                break
            shot_names.append(m.group(i)[1:])
        
        return {sequence_name: shot_names}

    def create_shot(self, name):
        if self._film.shots.has_mapped_name(name) == False:
            return self._film.shots.add_shot(
                name, self.get_kitsu_shot_names(name)
            )
        else:
            return self._film.shots[name]

    def import_shot_files(self, shots):
        for s in shots:
            # Get shot object
            print('ImportFiles :: {shot} - Import started'.format(shot=s))
            shot_data = self.shots_map[s]
            if shot_data.new_shot.get() == True:
                shot = self.create_shot(s)
                print('ImportFiles :: {shot} - Added to libreflow'.format(shot=s))
            else:
                shot = self._film.shots[s]
            
            # Import files
            for f in shot_data.files_map.mapped_items():
                task, target = f.file_target.get().split('/')
                
                # Task check
                if shot.tasks.has_mapped_name(task) == True:
                    file_map = shot.tasks[task].files
                    # File type check
                    if f.file_format.get() != '':
                        # For file
                        name, ext = target.rsplit('.', 1)
                        if file_map.has_file(name, ext):
                            f_object = file_map[f'{name}_{ext}']
                        else:
                            f_object = file_map.add_file(name, ext, tracked=True)
                        
                        r = self._create_revision(f_object)
                        revision_path = r.get_path()
                        revision_name = r.name()

                        if f.file_comment.get() != '':
                            r.comment.set(f.file_comment.get())
                        
                        print('ImportFiles :: {shot} - Importing {name} ({version})'.format(shot=s, name=target, version=revision_name))
                        shutil.copy2(f.file_path.get(), revision_path)
                        self._upload_revision(r)

                        #Delete file in files_map
                        shot_data.files_map.remove(f.name())
                    else:
                        # For folder
                        if file_map.has_folder(target):
                            f_object = file_map[target]

                            r = f_object.get_head_revision()
                            revision_path = r.get_path()

                            if os.path.exists(revision_path):
                                shutil.rmtree(revision_path)
                        else:
                            f_object = file_map.add_folder(target, tracked=True, default_path_format='{sequence}_{shot}/' + target)
                        
                        r = self._create_revision(f_object)
                        revision_path = r.get_path()
                        revision_name = r.name()

                        if f.file_comment.get() != '':
                            r.comment.set(f.file_comment.get())
                        
                        print('ImportFiles :: {shot} - Importing {name} ({version})'.format(shot=s, name=target, version=revision_name))
                        shutil.copytree(f.file_path.get(), revision_path)
                        self._upload_revision(r)

                        #Delete file in files_map
                        shot_data.files_map.remove(f.name())
                else:
                    print('ImportFiles [ERROR] :: {shot} - Can\'t find the task {task} for {name}. Import for this file is aborded'.format(shot=s, task=task, name=target))
                    continue
            
            if self.change_kitsu_status.get() == True:
                self.update_statutes(s)
            
            # if all files are imported, delete shot in shot_map of importfiles
            if shot_data.files_map.mapped_items() == []:
                self.shots_map.remove(s)
        
        print('ImportFiles :: Import complete!')

    def _create_revision(self, f):
        r = f.add_revision()
        revision_path = r.get_path()
        f.last_revision_oid.set(r.oid())
        os.makedirs(os.path.dirname(revision_path), exist_ok=True)

        return r

    def _upload_revision(self, revision):
        current_site = self.root().project().get_current_site()
        job = current_site.get_queue().submit_job(
            job_type='Upload',
            init_status='WAITING',
            emitter_oid=revision.oid(),
            user=self.root().project().get_user_name(),
            studio=current_site.name(),
        )

        self.root().project().get_sync_manager().process(job)

    def _update_kitsu_status(self, target):
        kitsu_api = self.root().project().kitsu_api()

        sequence_name, shot_name = target.get().split('_')
        sequence_name = sequence_name + '_'

        target_status = self.target_kitsu_task.get()
        task_name = target_status['name']
        task_status = target_status['status']

        kitsu_api.set_shot_task_status(sequence_name, shot_name, task_name, task_status)

    def _fill_ui(self, ui):
        ui['custom_page'] = 'libreflow.rouge.ui.ImportFilesWidget'


class Film(BaseFilm):
    
    shots = flow.Child(Shots).ui(
        expanded=True,
        show_filter=True,
    )

    sequences = flow.Child(flow.Object).ui(hidden=True)

    send_for_validation = flow.Child(SendForValidation)

    import_files = flow.Child(ImportFiles)



from kabaret.app.ui.gui.widgets.flow.flow_view import QtCore, QtGui, QtWidgets
from kabaret.app import resources

from libreflow.baseflow.runners import CHOICES_ICONS
from ...resources.icons import gui


class LabelIcon(QtWidgets.QLabel):

    def __init__(self, icon=None):
        QtWidgets.QLabel.__init__(self, '')

        icon = QtGui.QIcon(resources.get_icon(icon))
        pixmap = icon.pixmap(QtCore.QSize(16, 16))
        self.setPixmap(pixmap)
        self.setAlignment(QtCore.Qt.AlignVCenter)


class LineEdit(QtWidgets.QLineEdit):

    def __init__(self, value, dialog_widget):
        QtWidgets.QLineEdit.__init__(self)
        self.dialog_widget = dialog_widget
    
        if value != None:
            self.setText(value)

        self.editingFinished.connect(self.on_text_finish_edit)
         
    def on_text_finish_edit(self):
        if not self.text():
            self.setProperty('error', True)
            self.setStyleSheet('''
                QLineEdit {
                    border-color: red;
                }
                QToolTip {
                    background-color: #ffaaaa;
                    color: black;
                    border-color: red;
                }
            ''')
            error = '!!!\nERROR: Comment must not be empty.'
            self.setToolTip(error)
            return self.dialog_widget.refresh_buttons()
        
        self.setProperty('error', False)
        self.setStyleSheet('')
        self.setToolTip('')
        self.dialog_widget.refresh_buttons()


class EditFileComment(QtWidgets.QDialog):

    def __init__(self, tree, item):
        super(EditFileComment, self).__init__(tree)
        self.tree = tree
        self.custom_widget = tree.parent()
        self.item = item

        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setFixedSize(500, 0)

        self.lo = QtWidgets.QVBoxLayout(self)
        self.lo.setContentsMargins(20,20,20,20)

        palette = self.palette()
        palette.setColor(QtGui.QPalette.Base, palette.color(QtGui.QPalette.Window))
        self.setPalette(palette)

        self.content_lo = QtWidgets.QGridLayout()

        self.content_lo.addWidget(LabelIcon(('icons.flow', 'input')), 0, 0, QtCore.Qt.AlignVCenter)
        self.content_lo.addWidget(QtWidgets.QLabel('Comment'), 0, 1, QtCore.Qt.AlignVCenter)
        self.comment = LineEdit(self.item.toolTip(1), self)
        self.content_lo.addWidget(self.comment, 0, 2, QtCore.Qt.AlignVCenter)

        self.content_lo.setAlignment(QtCore.Qt.AlignTop)

        # Buttons
        self.button_lo = QtWidgets.QHBoxLayout()
        self.button_lo.addStretch()

        self.button_action = QtWidgets.QPushButton('Add')
        if self.comment.text() != '':
            self.button_action.setText('Edit')
        else:
            self.button_action.setEnabled(False)
        self.button_cancel = QtWidgets.QPushButton('Cancel')

        self.button_action.clicked.connect(self._on_action_button_clicked)
        self.button_cancel.clicked.connect(self._on_cancel_button_clicked)

        self.button_cancel.setAutoDefault(False)
        
        self.button_lo.addWidget(self.button_action)
        self.button_lo.addWidget(self.button_cancel)

        self.lo.addLayout(self.content_lo)
        self.lo.addLayout(self.button_lo)
        
        self.setLayout(self.lo)

    def sizeHint(self):
        return QtCore.QSize(500, 0)

    def refresh_buttons(self):
        for i in reversed(range(self.content_lo.count())):
            widget = self.content_lo.itemAt(i).widget()
            if widget:
                if widget.property('error') == True:
                    return self.button_action.setEnabled(False)
        
        return self.button_action.setEnabled(True)
    
    def _on_action_button_clicked(self):
        self.custom_widget.session.cmds.Flow.set_value(self.item.shot_file + '/file_comment', self.comment.text())
        self.tree.refresh()
        self.close()

    def _on_cancel_button_clicked(self):
        self.close()


class FileItem(QtWidgets.QTreeWidgetItem):

    def __init__(self, tree, item, shot_file):
        super(FileItem, self).__init__(item)
        self.tree = tree
        self.custom_widget = tree.parent()
        self.shot_file = shot_file

        self.refresh()

    def buttons(self):
        widget = QtWidgets.QWidget()
        lo = QtWidgets.QHBoxLayout(widget)
        lo.setContentsMargins(0, 0, 0, 0)
        lo.setSpacing(0)

        button_edit_comment = QtWidgets.QPushButton(
            QtGui.QIcon(resources.get_icon(('icons.gui', 'comment-lines'))), ''
        )
        button_edit_comment.setStyleSheet('qproperty-iconSize: 13px; padding: 2px;')
        button_edit_comment.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        button_edit_comment.clicked.connect(lambda checked=False, x=self: self._on_comment_button_clicked(x))
            
        lo.addStretch()
        lo.addWidget(button_edit_comment)

        return widget

    def refresh(self):
        self.setText(0, self.custom_widget.session.cmds.Flow.get_value(self.shot_file + '/file_name'))
        self.setIcon(0, self.get_file_icon(self.custom_widget.session.cmds.Flow.get_value(self.shot_file + '/file_format')))

        comment = self.custom_widget.session.cmds.Flow.get_value(self.shot_file + '/file_comment')
        self.setIcon(1, QtGui.QIcon())
        self.setToolTip(1, '')
        if comment != '':
            self.setIcon(1, self.get_icon(('icons.gui', 'speech-bubbles-comment-option')))
            self.setToolTip(1, comment)
        
        self.setText(2, self.custom_widget.session.cmds.Flow.get_value(self.shot_file + '/file_target'))
        self.setText(3, self.custom_widget.session.cmds.Flow.get_value(self.shot_file + '/rev_version'))

        self.tree.setItemWidget(self, 4, self.buttons())

    @staticmethod
    def get_file_icon(extension=None):
        if extension == '':
            return QtGui.QIcon(resources.get_icon(('icons.gui', 'folder-white-shape')))
        else:
            return QtGui.QIcon(resources.get_icon(CHOICES_ICONS.get(extension, ('icons.gui', 'text-file-1'))))

    @staticmethod
    def get_icon(icon_ref):
        return QtGui.QIcon(resources.get_icon(icon_ref))

    def _on_comment_button_clicked(self, item):
        dialog = EditFileComment(self.tree, self)
        dialog.exec()


class ShotListItem(QtWidgets.QTreeWidgetItem):

    def __init__(self, tree, shot):
        super(ShotListItem, self).__init__(tree)
        self.tree = tree
        self.custom_widget = tree.parent()
        self.shot = shot
        self.status = self.custom_widget.session.cmds.Flow.get_value(self.shot + '/shot_status')

        self.refresh()
    
    def refresh(self):
        files = self.custom_widget.get_shot_files(self.shot)
        shot_name = self.custom_widget.session.cmds.Flow.get_value(self.shot + '/shot_name')
        if files == []:
            self.custom_widget.remove_shot(shot_name)
            return self.tree.takeTopLevelItem(self.tree.indexFromItem(self).row())

        for shot_file in files:
            FileItem(self.tree, self, shot_file)

        self.setFlags(QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsUserCheckable)
        self.setCheckState(0, QtCore.Qt.Checked)
        self.setExpanded(True)
        
        self.setText(0, shot_name)
        if self.new_shot() == True:
            self.setIcon(1, self.get_icon(('icons.gui', 'plus-black-symbol')))
            self.setToolTip(1, 'Shot will be created in libreflow')
        
        self.paint()

    def new_shot(self):
        return self.custom_widget.session.cmds.Flow.get_value(self.shot + '/new_shot')

    def paint(self):
        for i in range(self.treeWidget().header().count()):
            brush = QtGui.QBrush(QtGui.QColor(70, 70, 70))
            self.setBackground(i, brush)
    
    @staticmethod
    def get_icon(icon_ref):
        return QtGui.QIcon(resources.get_icon(icon_ref))


class ShotList(QtWidgets.QTreeWidget):

    def __init__(self, custom_widget):
        super(ShotList, self).__init__(custom_widget)
        self._custom_widget = custom_widget
        self.overwrite = []

        self.setStyleSheet("QTreeWidget::item { height: 25px }")

        self.setAcceptDrops(True)
        self.setHeaderLabels(self.get_columns())
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        
        self.refresh(True)

        self.header().setStretchLastSection(True)
        self.setColumnWidth(0, 200)
        self.setColumnWidth(1, 170)
        self.setColumnWidth(2, 240)
        self.setColumnWidth(3, 60)

        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._on_context_menu)

    def get_columns(self):
        return ('Source file', '', 'Target', 'Version', '')
    
    def sizeHint(self):
        return QtCore.QSize(800, 500)
    
    def refresh(self, init=None):
        self.blockSignals(True)
        self.clear()

        shots = self._custom_widget.get_shots()

        for shot in shots:
            ShotListItem(self, shot)

        if init is None:
            if self.topLevelItemCount() == 0:
                self._custom_widget.checkbox_selectall.setCheckState(QtCore.Qt.Unchecked)
                self._custom_widget.checkbox_selectall.setEnabled(False)
                self._custom_widget.button_import.setEnabled(False)
                self._custom_widget.dragdrop.show()
                self._custom_widget.shot_list.hide()
            else:
                self._custom_widget.checkbox_selectall.setEnabled(True)
                self._custom_widget.button_import.setEnabled(True)
            self._custom_widget.refresh_shots_count()
        self.blockSignals(False)

    def get_shots_count(self):
        count = self.topLevelItemCount()
        if count > 1:
            return str(count) + " shots"
        if count == 1:
            return str(count) + " shot"
        return str(0) + " shots"

    @staticmethod
    def get_icon(icon_ref):
        return QtGui.QIcon(resources.get_icon(icon_ref))

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls:
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        if event.mimeData().hasUrls:
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        if event.mimeData().hasUrls:
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
            links = []
            for url in event.mimeData().urls():
                links.append(str(url.toLocalFile()))
            self._custom_widget.add_files(links)
        else:
            event.ignore()

    def _on_remove_comment_action_clicked(self, item):
        self._custom_widget.session.cmds.Flow.set_value(item.shot_file + '/file_comment', '')
        item.refresh()

    def _on_remove_file_action_clicked(self, item):
        name = self._custom_widget.session.cmds.Flow.call(
            item.shot_file, 'name', {}, {}
        )
        self._custom_widget.session.cmds.Flow.call(
            item.parent().shot + '/files_map', 'remove', [name], {}
        )
        self.refresh()

    def _on_remove_all_files_action_clicked(self, item):
        self._custom_widget.session.cmds.Flow.call(
            item.shot + '/files_map', 'clear', {}, {}
        )
        self.refresh()

    def _on_context_menu(self, event):
        item = self.itemAt(event)
        column = self.currentColumn()

        if item is None:
            return

        context_menu = QtWidgets.QMenu(self)

        if item.childCount() != 0:
            remove_all_files = context_menu.addAction(self.get_icon(('icons.gui', 'remove-symbol')), 'Remove Shot')
        
            remove_all_files.triggered.connect(
                lambda checked=False, x=item: self._on_remove_all_files_action_clicked(x)
            )
        else:
            if item.toolTip(1) != '':
                remove_comment = context_menu.addAction(self.get_icon(('icons.gui', 'speech-bubbles-comment-option')), 'Remove Comment')
                remove_comment.triggered.connect(
                    lambda checked=False, x=item: self._on_remove_comment_action_clicked(x)
                )

            if item.parent().status == 'init':
                optional_init = self._custom_widget.session.cmds.Flow.get_value(item.shot_file + '/optional_init')
                if optional_init == True:
                    remove_file = context_menu.addAction(self.get_icon(('icons.gui', 'remove-symbol')), 'Remove File')
                    remove_file.triggered.connect(
                        lambda checked=False, x=item: self._on_remove_file_action_clicked(x)
                    )
            if item.parent().status == 'rtk':
                optional_new_rev = self._custom_widget.session.cmds.Flow.get_value(item.shot_file + '/optional_new_rev')
                if optional_new_rev == True:
                    remove_file = context_menu.addAction(self.get_icon(('icons.gui', 'remove-symbol')), 'Remove File')
                    remove_file.triggered.connect(
                        lambda checked=False, x=item: self._on_remove_file_action_clicked(x)
                    )

        if context_menu.isEmpty() != True:
            context_menu.exec_(self.mapToGlobal(event))


from kabaret import flow
from libreflow.baseflow.site import WorkingSite as BaseWorkingSite
from libreflow.utils.flow.values import MultiOSParam


class WorkingSite(BaseWorkingSite):

    with flow.group('User local workspace'):
        enable_user_workspace = flow.BoolParam(False)
        user_working_dir      = MultiOSParam()
    with flow.group('Package'):
        delivery_dir          = MultiOSParam()

    visible_tasks         = flow.OrderedStringSetParam()



from kabaret import flow
from kabaret.flow_entities.entities import Property
from libreflow.baseflow.shot import (
    ShotCollection,
    Shot as BaseShot,
    Sequence as BaseSequence
)

from .task import Tasks


class Shot(BaseShot):
    '''
    This shot class holds a `kitsu_shots` property which
    provides the Kitsu shot entities bound to this object.

    This property is suited for binding multiple Kitsu shots
    to a single flow shot.
    Its value must be a dictionnary mapping the sequence names
    with the lists of bound shot names:
        {<sequence_name>: [<shot_name>, ...]}
    '''

    kitsu_shots = Property().ui(hidden=True)
    
    tasks = flow.Child(Tasks).ui(
        expanded=True
    )

    @classmethod
    def get_source_display(cls, oid):
        split = oid.split('/')
        return f'{split[3]} · {split[5]}'

    def ensure_tasks(self):
        """
        Creates the tasks of this shot based on the task
        templates of the project, skipping any existing task.
        """
        mgr = self.root().project().get_task_manager()

        for dt in mgr.default_tasks.mapped_items():
            if not self.tasks.has_mapped_name(dt.name()) and not dt.optional.get():
                t = self.tasks.add(dt.name())
                t.enabled.set(dt.enabled.get())
        
        self.tasks.touch()

    def get_default_contextual_edits(self, context_name):
        if context_name == 'settings':
            sequence, shot = self.name().split('_', maxsplit=1)
            return dict(sequence=sequence, shot=shot)
    
    def set_task_status(self, task_name, task_status, comment=''):
        kitsu_shots = self.kitsu_shots.get()

        if not kitsu_shots:
            raise Exception(
                f'Kitsu shots associated to {self.oid()} undefined'
            )
        
        kitsu = self.root().project().kitsu_api()

        for sequence_name, shot_names in kitsu_shots.items():
            for shot_name in shot_names:
                kitsu.set_shot_task_status(
                    sequence_name, shot_name, task_name, task_status, comment
                )
    
    def get_task_status(self, task_name):
        kitsu_shots = self.kitsu_shots.get()

        if not kitsu_shots:
            raise Exception(
                f'Kitsu shots associated to {self.oid()} undefined'
            )
        
        sequence_name, shot_names = list(kitsu_shots.items())[0]
        
        kitsu = self.root().project().kitsu_api()
        return kitsu.get_shot_task_status_name(
            sequence_name, shot_names[0], task_name
        )


class Shots(ShotCollection):
    
    def add_shot(self, name, kitsu_shots):
        s = self.add(name)
        s.kitsu_shots.set(kitsu_shots)
        s.ensure_tasks()

        return s



import os
import shutil
import pathlib

from kabaret import flow
from libreflow.baseflow.file import (
    TrackedFile                   as BaseTrackedFile,
    TrackedFolder                 as BaseTrackedFolder,
    Revision                      as BaseRevision,
    TrackedFolderRevision         as BaseTrackedFolderRevision,
    FileSystemMap                 as BaseFileSystemMap,
    CreateWorkingCopyAction       as BaseCreateWorkingCopyAction,
    CreateWorkingCopyFromRevision as BaseCreateWorkingCopyFromRevision,
    PublishFileAction             as BasePublishFileAction,
    RevisionsWorkingCopiesChoiceValue,
)
from libreflow.utils.flow import get_contextual_dict


class Revision(BaseRevision):
    
    def get_path(self, relative=False):
        path = self.path.get()

        if path == '' or path is None:
            raise Exception((
                f'Cannot call get_path() on revision {self.oid()} '
                 'since its path is undefined.'
            ))

        if not relative:
            root = self.root().project().get_root(
                user_workspace=self.working_copy.get()
            )
            path = os.path.join(root, path)
        
        return path


class TrackedFolderRevision(BaseTrackedFolderRevision):
    pass


class RevisionsWorkingCopies(RevisionsWorkingCopiesChoiceValue):

    _files = flow.Parent(3)
    _shot = flow.Parent(6)

    def _fill_ui(self, ui):
        ui['hidden'] = (
            self._shot.get_task_status('L&S ANIMATION') == 'K APPROVED'
            or not self._files.has_folder('pack')
            or self._files['pack'].get_head_revision() is None
            or self._files['pack'].get_head_revision().get_sync_status() != 'Available'
        )


class CreateWorkingCopyFromRevision(BaseCreateWorkingCopyFromRevision):

    _file = flow.Parent(4)

    def allow_context(self, context):
        user = self.root().project().get_user()
        is_head = True
        if user.all_working_copy.get() == False:
            head = self._file.get_head_revision().name()
            if self._revision.name() != head:
                is_head = False

        if user.preferences.create_working_copies.get() == True:
            return (
                context
                and self._revision._file.editable()
                and self._revision.name() != self.root().project().get_user_name()
                and is_head
            )
        else:
            return (
                context
                and self._revision._file.editable()
                and not self._revision.is_working_copy()
                and is_head
            )


class CreateWorkingCopyAction(BaseCreateWorkingCopyAction):

    _file = flow.Parent()
    _files = flow.Parent(2)
    _task = flow.Parent(3)

    from_revision = flow.Param(None, RevisionsWorkingCopies).ui(
        label='Revision'
    )

    _shot = flow.Parent(5)

    def get_buttons(self):
        status = self._shot.get_task_status('L&S ANIMATION')
        if status == 'K APPROVED':
            self.message.set((
                '<h3>Create a working copy</h3>'
                '<font color=#D5000D>You can\'t create a working '
                'copy since the status of the <b>L&S ANIMATION</b> '
                'task is <b>APPROVED</b>.</font>'
            ))
            return ['Cancel']
        elif (
            not self._files.has_folder('pack')
            or self._files['pack'].get_head_revision() is None
            or self._files['pack'].get_head_revision().get_sync_status() != 'Available'
        ):
            self.message.set((
                '<h3>Create a working copy</h3>'
                '<font color=#D5000D>You can\'t create a working '
                'copy since no <b>pack</b> folder has been prepared '
                'for this shot.</font>'
            ))
            return ['Cancel']
        else:
            return super(CreateWorkingCopyAction, self).get_buttons()
    
    def run(self, button):
        super(CreateWorkingCopyAction, self).run(button)

        if button == 'Create' or button == 'Create from scratch':
            # Copy pack into user workspace if not copied yet
            self._ensure_pack_in_workspace()

            # Update Kitsu status of the current task
            user_name = self.root().project().get_user_name()

            if self._shot.get_task_status('L&S KEY FRAME') != 'K APPROVED':
                target_task = 'L&S KEY FRAME'
            else:
                target_task = 'L&S ANIMATION'

            self._shot.set_task_status(
                target_task,
                'E Work In Progress',
                comment=(
                    f"**{user_name}** has created a working copy on "
                    f"'{self._task.name()}/{self._file.display_name.get()}'."
                )
            )
    
    def _ensure_pack_in_workspace(self):
        # Retrieve the path of the pack in the user workspace
        pack = self._files['pack']
        working_dir = self.root().project().get_root(user_workspace=True)
        pack_relpath = pack.get_head_revision().get_path(relative=True)
        path_in_workspace = os.path.join(working_dir, pack_relpath)

        if not os.path.isdir(path_in_workspace):
            path_on_site = os.path.join(self.root().project().get_root(), pack_relpath)
            shutil.copytree(path_on_site, path_in_workspace)


class KitsuTaskNames(flow.values.SessionValue):

    DEFAULT_EDITOR = 'choice'

    STRICT_CHOICES = False

    _task = flow.Parent(4)

    def choices(self):
        kitsu = self.root().project().kitsu_api()
        tm = self.root().project().get_task_manager()
        subtasks = tm.get_subtasks(self._task.name())

        return [
            kitsu.get_task_name(self._task.name(), st)
            for st in subtasks
        ]
    
    def revert_to_default(self):
        current = self._task.current_subtask.get()
        kitsu = self.root().project().kitsu_api()
        if current:
            kitsu_task = kitsu.get_task_name(self._task.name(), current)
            self.set(kitsu_task)
        else:
            names = self.choices()
            if names:
                self.set(names[0])


class KitsuTargetStatus(flow.values.SessionValue):

    DEFAULT_EDITOR = 'choice'

    _statutes = {
        'WIP': 'E Work In Progress',
        'To check': 'F To CHECK ',
    }

    def choices(self):
        return ['WIP', 'To check']
    
    def get_kitsu_status(self):
        return self._statutes[self._value]


class PublishFileAction(BasePublishFileAction):

    keep_editing = flow.SessionParam(False).ui(
        editor='bool',
        hidden=True,
    )
    upload_after_publish = flow.SessionParam(False).ui(
        editor='bool',
        hidden=True,
    )
    status = flow.SessionParam('WIP', KitsuTargetStatus).ui(
        label='Kitsu status'
    )

    _file = flow.Parent()
    _task = flow.Parent(3)
    _shot = flow.Parent(5)

    def check_default_values(self):
        self.comment.apply_preset()
    
    def update_presets(self):
        self.comment.update_preset()

    def run(self, button):
        if button == 'Cancel':
            return
        
        # Keep working copy when Kitsu status is set to WIP
        self.keep_editing.set(
            self.status.get() == 'WIP'
        )

        super(PublishFileAction, self).run(button)
        
        user_name = self.root().project().get_user_name()

        key_frame_status = self._shot.get_task_status('L&S KEY FRAME')
        target_task = 'L&S KEY FRAME'

        if key_frame_status == 'K APPROVED':
            target_task = 'L&S ANIMATION'

        self._shot.set_task_status(
            target_task,
            self.status.get_kitsu_status(),
            comment=(
                f"**{user_name}** has published version **{self._file.get_head_revision().name()}**.\n"
                f"> {self.comment.get()}\n\n"
                f"*{self._task.oid()}*"
            )
        )


class SendFileForValidation(flow.Action):

    _file = flow.Parent()
    _map  = flow.Parent(2)
    _shot = flow.Parent(5)
    _film = flow.Parent(7)

    kitsu_target_task   = flow.SessionParam(value_type=KitsuTaskNames).watched()
    send_pack           = flow.SessionParam(False).ui(editor='bool')
    kitsu_source_status = flow.Param('H TO SEND').ui(hidden=True)
    kitsu_target_status = flow.Param('I Waiting For Approval').ui(hidden=True)

    def allow_context(self, context):
        return (
            context
            and self._file.name() == 'working_file_plas'
            and not self._file.is_empty()
        )
    
    def needs_dialog(self):
        self.kitsu_target_task.revert_to_default()
        return True

    def get_buttons(self):
        return ['Confirm', 'Cancel']
    
    def run(self, button):
        if button == 'Cancel':
            return
        elif not self._is_status_valid():
            return self.get_result(close=False)
        
        send_for_validation = self._film.send_for_validation
        delivery_dir = send_for_validation.get_delivery_dir()

        take_name, source_revision_name = send_for_validation.ensure_take(
            self._shot.name(),
            'working_file_plas',
            self.kitsu_target_task.get(),
            replace_take=False
        )
        send_for_validation.create_shot_package(
            self._shot.name(),
            self.kitsu_target_task.get(),
            take_name,
            include_pack=self.send_pack.get(),
            delivery_dir=delivery_dir
        )
        send_for_validation.update_kitsu_status(
            self._shot.name(),
            self.kitsu_target_task.get(),
            self.kitsu_target_status.get(),
            take_name, source_revision_name
        )
    
    def child_value_changed(self, child_value):
        if child_value is self.kitsu_target_task:
            self._check_kitsu_status()
            self.send_pack.set((
                self.kitsu_target_task.get() == 'L&S KEY FRAME'
                and (
                    not self._map.has_file('takes_fix', 'plas')
                    or not self._map['takes_fix_plas'].has_revision('t1')
                )
            ))
    
    def _is_status_valid(self):
        kitsu_task_name = self.kitsu_target_task.get()
        source_status = self.kitsu_source_status.get()

        return self._shot.get_task_status(kitsu_task_name) == source_status
    
    def _check_kitsu_status(self):
        if not self._is_status_valid():
            self.message.set((
                '<font color=#D5000D>'
                'You cannot send this file for validation since '
                f'its Kitsu status is not <b>{self.kitsu_source_status.get()}</b>.'
                '</font>'
            ))
        else:
            self.message.set('')


class TrackedFile(BaseTrackedFile):
    
    to_validate = flow.Child(SendFileForValidation)


class TrackedFolder(BaseTrackedFolder):
    pass


class FileSystemMap(BaseFileSystemMap):
    
    def add_file(self, name, extension, display_name=None, base_name=None, tracked=False, default_path_format=None):
        if default_path_format is None:
            default_path_format = get_contextual_dict(self, 'settings').get(
                'path_format', None
            )
        return super(FileSystemMap, self).add_file(name, extension, display_name, base_name, tracked, default_path_format)

    def add_folder(self, name, display_name=None, base_name=None, tracked=False, default_path_format=None):
        if default_path_format is None:
            default_path_format = get_contextual_dict(self, 'settings').get(
                'path_format', None
            )
        return super(FileSystemMap, self).add_folder(name, display_name, base_name, tracked, default_path_format)



from enum import Enum
from kabaret.app.ui.gui.widgets.flow.flow_view import QtCore, QtGui, QtWidgets, CustomPageWidget
from kabaret.app import resources
from libreflow.resources.icons import libreflow as _

from ..flow.film import ShotState


class ShotSelectWidget(QtWidgets.QWidget):

    def __init__(self, shot_item, shot_name):
        super(ShotSelectWidget, self).__init__()
        self._shot_item = shot_item
        self._shot_name_label = QtWidgets.QLabel(shot_name)
        self.checkbox_select = QtWidgets.QCheckBox()

        hlo = QtWidgets.QHBoxLayout()
        hlo.addWidget(self.checkbox_select)
        hlo.addWidget(self._shot_name_label)
        hlo.setContentsMargins(0, 0, 0, 0)
        hlo.setSpacing(0)

        self.setLayout(hlo)
    
    def set_label_color(self, color):
        self.setStyleSheet('QLabel { color : %s; }' % color)
    
    def sizeHint(self):
        return QtCore.QSize(150, 30)


class ShotToSendItem(QtWidgets.QTreeWidgetItem):

    def __init__(self, tree, oid, custom_widget, session):
        super(ShotToSendItem, self).__init__(tree)
        self._custom_widget = custom_widget
        self._session = session
        self._oid = oid
        self._state = None

        self._select_widget = ShotSelectWidget(self, self.get_shot_name())
        self._button_choose_take = QtWidgets.QPushButton('...')
        self._checkbox_pack = QtWidgets.QCheckBox()
        self.treeWidget().setItemWidget(self, 0, self._select_widget)
        self.treeWidget().setItemWidget(self, 2, self._checkbox_pack)
        self.setTextAlignment(1, QtCore.Qt.AlignCenter)
        
        self.initialise()

        # Install widget callbacks after their initialisation
        self._select_widget.checkbox_select.stateChanged.connect(self._on_checkbox_select_state_changed)
        self._checkbox_pack.stateChanged.connect(self._on_checkbox_pack_state_changed)
        self._button_choose_take.clicked.connect(self._on_button_choose_clicked)
    
    def initialise(self):
        self.setText(2, '')
        init_state = self._session.cmds.Flow.call(self._oid, 'get_init_state', [], {})

        if init_state == ShotState.MISSING_FILE:
            self._get_checkbox_select().setCheckState(QtCore.Qt.Unchecked)
            self._get_checkbox_pack().setCheckState(QtCore.Qt.Unchecked)
            self.set_selected(False)
            self.set_send_pack(False)
            self.set_new_take(False)
        elif init_state == ShotState.FIRST_DELIVERY:
            self._get_checkbox_select().setCheckState(QtCore.Qt.Checked)
            self._get_checkbox_pack().setCheckState(QtCore.Qt.Checked)
            self.set_selected(True)
            self.set_send_pack(True)
            self.set_new_take(True)
        elif init_state == ShotState.NEW_TAKE_PACK_UNSELECTED:
            self._get_checkbox_select().setCheckState(QtCore.Qt.Checked)
            self._get_checkbox_pack().setCheckState(QtCore.Qt.Unchecked)
            self.set_selected(True)
            self.set_send_pack(False)
            self.set_new_take(True)
        elif init_state == ShotState.ALREADY_DELIVERED:
            self._get_checkbox_select().setCheckState(QtCore.Qt.Unchecked)
            self._get_checkbox_pack().setCheckState(QtCore.Qt.Unchecked)
            self.set_selected(False)
            self.set_send_pack(False)
            self.set_new_take(False)
        elif init_state == ShotState.ALREADY_RECEIVED:
            self._get_checkbox_select().setCheckState(QtCore.Qt.Unchecked)
            self._get_checkbox_pack().setCheckState(QtCore.Qt.Unchecked)
            self.set_selected(False)
            self.set_send_pack(False)
            self.set_new_take(True)
        
        self.update_state(init_state)
    
    def get_shot_name(self):
        return self._session.cmds.Flow.call(self._oid, 'name', [], {})
    
    def get_target_take_name(self):
        return self._session.cmds.Flow.get_value(self._oid+'/target_take')
    
    def get_state(self):
        return self._state
    
    def update_state(self, state):
        self._state = state
        self._activate(self._state)
    
    def set_selected(self, b):
        self._session.cmds.Flow.set_value(self._oid+'/selected', b)
    
    def set_new_take(self, b):
        self._session.cmds.Flow.set_value(self._oid+'/new_take', b)
    
    def set_send_pack(self, b):
        self._session.cmds.Flow.set_value(self._oid+'/send_pack', b)
    
    def _activate(self, state):
        if state == ShotState.MISSING_FILE:
            self._get_checkbox_select().setEnabled(False)
            self._get_checkbox_pack().setEnabled(False)
            
            self._select_widget.set_label_color('#d5000d')
            self.setForeground(3, QtGui.QBrush(QtGui.QColor('#d5000d')))
            
            self.setText(1, '')
            self.setText(3, 'Missing working file')
            self.treeWidget().setItemWidget(self, 1, QtWidgets.QWidget())
        elif state == ShotState.FIRST_DELIVERY:
            self._get_checkbox_select().setEnabled(True)
            self._get_checkbox_pack().setEnabled(False)

            self._select_widget.set_label_color('#b9c2c8')
            self.setForeground(1, QtGui.QBrush(QtGui.QColor('#b9c2c8')))
            self.setForeground(3, QtGui.QBrush(QtGui.QColor('#3b9800')))
            
            self.setText(1, self.get_target_take_name())
            self.setText(3, 'First delivery')
            self.treeWidget().setItemWidget(self, 1, QtWidgets.QWidget())
        elif state == ShotState.NEW_TAKE_PACK_SELECTED:
            self._get_checkbox_select().setEnabled(True)
            self._get_checkbox_pack().setEnabled(True)

            self._select_widget.set_label_color('#b9c2c8')
            self.setForeground(1, QtGui.QBrush(QtGui.QColor('#b9c2c8')))
            self.setForeground(3, QtGui.QBrush(QtGui.QColor('#b9c2c8')))
            
            self.setText(1, self.get_target_take_name())
            self.setText(3, 'Pack will be sent again')
            self.treeWidget().setItemWidget(self, 1, QtWidgets.QWidget())
        elif state == ShotState.NEW_TAKE_PACK_UNSELECTED:
            self._get_checkbox_select().setEnabled(True)
            self._get_checkbox_pack().setEnabled(True)

            self._select_widget.set_label_color('#b9c2c8')
            self.setForeground(1, QtGui.QBrush(QtGui.QColor('#b9c2c8')))
            
            self.setText(1, self.get_target_take_name())
            self.setText(3, '')
            self.treeWidget().setItemWidget(self, 1, QtWidgets.QWidget())
        elif state == ShotState.RESEND_TAKE_PACK_SELECTED:
            self._get_checkbox_select().setEnabled(True)
            self._get_checkbox_pack().setEnabled(True)

            self._select_widget.set_label_color('#b9c2c8')
            self.setForeground(1, QtGui.QBrush(QtGui.QColor('#d5000d')))
            self.setForeground(3, QtGui.QBrush(QtGui.QColor('#b9c2c8')))
            
            self.setText(1, self.get_target_take_name())
            self.setText(3, 'Pack will be sent again')
            self.treeWidget().setItemWidget(self, 1, QtWidgets.QWidget())
        elif state == ShotState.RESEND_TAKE_PACK_UNSELECTED:
            self._get_checkbox_select().setEnabled(True)
            self._get_checkbox_pack().setEnabled(True)

            self._select_widget.set_label_color('#b9c2c8')
            self.setForeground(1, QtGui.QBrush(QtGui.QColor('#d5000d')))
            
            self.setText(1, self.get_target_take_name())
            self.setText(3, '')
            self.treeWidget().setItemWidget(self, 1, QtWidgets.QWidget())
        elif state == ShotState.ALREADY_DELIVERED:
            self._get_checkbox_select().setEnabled(False)
            self._get_checkbox_pack().setEnabled(False)

            self._select_widget.set_label_color('#d66700')
            self.setForeground(3, QtGui.QBrush(QtGui.QColor('#d66700')))
            
            self.setText(1, '')
            self.setText(3, 'Already delivered')
            self.treeWidget().setItemWidget(self, 1, self._button_choose_take)
            
            self._get_checkbox_select().setEnabled(False)
        elif state == ShotState.ALREADY_RECEIVED:
            self._get_checkbox_select().setEnabled(False)
            self._get_checkbox_pack().setEnabled(False)

            self._select_widget.set_label_color('#d66700')
            self.setForeground(3, QtGui.QBrush(QtGui.QColor('#d66700')))
            
            self.setText(1, '')
            self.setText(3, 'Already received')
            self.treeWidget().setItemWidget(self, 1, self._button_choose_take)
    
    def _get_checkbox_select(self):
        return self._select_widget.checkbox_select
    
    def _get_checkbox_pack(self):
        return self._checkbox_pack
    
    def _on_checkbox_select_state_changed(self, checkstate):
        if self._state == ShotState.FIRST_DELIVERY:
            self.set_selected(checkstate == QtCore.Qt.Checked)
        elif self._state == ShotState.NEW_TAKE_PACK_SELECTED:
            self.set_selected(checkstate == QtCore.Qt.Checked)
        elif self._state == ShotState.NEW_TAKE_PACK_UNSELECTED:
            self.set_selected(checkstate == QtCore.Qt.Checked)
        elif self._state == ShotState.RESEND_TAKE_PACK_SELECTED:
            self.set_selected(checkstate == QtCore.Qt.Checked)
        elif self._state == ShotState.RESEND_TAKE_PACK_UNSELECTED:
            self.set_selected(checkstate == QtCore.Qt.Checked)
    
    def _on_checkbox_pack_state_changed(self, checkstate):
        if self._state == ShotState.NEW_TAKE_PACK_SELECTED:
            if checkstate == QtCore.Qt.Unchecked:
                self.update_state(ShotState.NEW_TAKE_PACK_UNSELECTED)
                self.set_send_pack(False)
        elif self._state == ShotState.NEW_TAKE_PACK_UNSELECTED:
            if checkstate == QtCore.Qt.Checked:
                self.update_state(ShotState.NEW_TAKE_PACK_SELECTED)
                self.set_send_pack(True)
        elif self._state == ShotState.RESEND_TAKE_PACK_SELECTED:
            if checkstate == QtCore.Qt.Unchecked:
                self.update_state(ShotState.NEW_TAKE_PACK_UNSELECTED)
                self.set_send_pack(False)
        elif self._state == ShotState.RESEND_TAKE_PACK_UNSELECTED:
            if checkstate == QtCore.Qt.Checked:
                self.update_state(ShotState.RESEND_TAKE_PACK_SELECTED)
                self.set_send_pack(True)
    
    def _on_button_choose_clicked(self, checked=False):
        if self._state == ShotState.ALREADY_DELIVERED:
            res = self._custom_widget.dialog_choose_take.show_dialog(self._state)

            if res == QtWidgets.QDialog.Accepted:
                new_take = self._custom_widget.dialog_choose_take.get_new_take()
                self.set_new_take(new_take)
                self._get_checkbox_select().setCheckState(QtCore.Qt.Checked)
                self.set_selected(True)

                if new_take:
                    self.update_state(ShotState.NEW_TAKE_PACK_UNSELECTED)
                else:
                    self.update_state(ShotState.RESEND_TAKE_PACK_UNSELECTED)
        elif self._state == ShotState.ALREADY_RECEIVED:
            res = self._custom_widget.dialog_choose_take.show_dialog(self._state)

            if res == QtWidgets.QDialog.Accepted:
                self.set_new_take(True)
                self.update_state(ShotState.NEW_TAKE_PACK_UNSELECTED)


class SelectAllWidget(QtWidgets.QWidget):

    def __init__(self):
        super(SelectAllWidget, self).__init__()
        self._label = QtWidgets.QLabel('Select all')
        self.checkbox = QtWidgets.QCheckBox()

        self.setStyleSheet('QLabel { font-weight : 600; }')

        hlo = QtWidgets.QHBoxLayout()
        hlo.addWidget(self.checkbox)
        hlo.addWidget(self._label)
        hlo.setContentsMargins(0, 0, 0, 0)
        hlo.setSpacing(0)

        self.setLayout(hlo)
    
    def sizeHint(self):
        return QtCore.QSize(60, 30)


class SelectAllItem(QtWidgets.QTreeWidgetItem):
    
    def __init__(self, tree):
        super(SelectAllItem, self).__init__(tree)
        self._widget_select = SelectAllWidget()
        self._widget_pack = SelectAllWidget()
        self.treeWidget().setItemWidget(self, 0, self._widget_select)
        self.treeWidget().setItemWidget(self, 2, self._widget_pack)
        
        for i in range(self.treeWidget().columnCount()):
            self.setBackground(i, QtGui.QBrush(QtGui.QColor('#311215')))
        
        self._widget_select.checkbox.stateChanged.connect(self._on_checkbox_select_state_changed)
        self._widget_pack.checkbox.stateChanged.connect(self._on_checkbox_pack_state_changed)
    
    def _on_checkbox_select_state_changed(self, checkstate):
        for i in range(1, self.treeWidget().topLevelItemCount()):
            item = self.treeWidget().topLevelItem(i)
            if (
                    item.get_state() != ShotState.MISSING_FILE
                and item.get_state() != ShotState.ALREADY_DELIVERED
                and item.get_state() != ShotState.ALREADY_RECEIVED
            ):
                item._get_checkbox_select().setCheckState(
                    QtCore.Qt.CheckState(checkstate)
                )

    def _on_checkbox_pack_state_changed(self, checkstate):
        for i in range(1, self.treeWidget().topLevelItemCount()):
            item = self.treeWidget().topLevelItem(i)
            if (
                    item.get_state() != ShotState.MISSING_FILE
                and item.get_state() != ShotState.FIRST_DELIVERY
                and item.get_state() != ShotState.ALREADY_DELIVERED
                and item.get_state() != ShotState.ALREADY_RECEIVED
            ):
                item._get_checkbox_pack().setCheckState(
                    QtCore.Qt.CheckState(checkstate)
                )


class ShotToSendList(QtWidgets.QTreeWidget):
    
    def __init__(self, custom_widget, session):
        super(ShotToSendList, self).__init__()
        self.custom_widget = custom_widget
        self.session = session

        self.setHeaderLabels(self.get_header_labels())
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)

        self.update()
        self.header().resizeSections(QtWidgets.QHeaderView.ResizeToContents)
    
    def get_header_labels(self):
        return ['Shots', 'Delivered take', 'Send pack', 'Info']
    
    def update(self):
        self.clear()
        SelectAllItem(self)

        oids = self.session.cmds.Flow.get_mapped_oids(
            self.custom_widget.oid+'/shots',
        )
        for oid in oids:
            ShotToSendItem(self, oid, self.custom_widget, self.session)


class ChooseTakeWidget(QtWidgets.QWidget):

    def __init__(self):
        super(ChooseTakeWidget, self).__init__()
        self._radiobutton_new_take = QtWidgets.QRadioButton('Create a new take')
        self._radiobutton_resend = QtWidgets.QRadioButton('Resend the last take')
        self._new_take_mandatory_label = QtWidgets.QLabel('Last take already received: create a new take ?')
        self.setStyleSheet('QLabel { color : #d66700; }')

        vlo = QtWidgets.QVBoxLayout()
        vlo.addWidget(self._radiobutton_new_take)
        vlo.addWidget(self._radiobutton_resend)
        vlo.addWidget(self._new_take_mandatory_label)

        vlo.setContentsMargins(0, 0, 0, 0)
        vlo.setSpacing(0)

        self.setLayout(vlo)
    
    def update(self, state):
        if state == ShotState.ALREADY_DELIVERED:
            self._radiobutton_new_take.setVisible(True)
            self._radiobutton_resend.setVisible(True)
            self._new_take_mandatory_label.setVisible(False)

            self._radiobutton_new_take.setChecked(True)
        elif state == ShotState.ALREADY_RECEIVED:
            self._radiobutton_new_take.setVisible(False)
            self._radiobutton_resend.setVisible(False)
            self._new_take_mandatory_label.setVisible(True)
            
            self._radiobutton_new_take.setChecked(True)


class ChooseTakeDialog(QtWidgets.QDialog):

    def __init__(self, parent):
        super(ChooseTakeDialog, self).__init__(parent)
        self._widget_choose_take = ChooseTakeWidget()
        self._button_confirm = QtWidgets.QPushButton('Confirm')
        self._button_cancel = QtWidgets.QPushButton('Cancel')

        glo = QtWidgets.QGridLayout()
        glo.addWidget(self._widget_choose_take, 0, 0, 1, 2)
        glo.addWidget(self._button_confirm, 1, 0)
        glo.addWidget(self._button_cancel, 1, 1)

        self.setLayout(glo)

        self.setWindowTitle('')
        self.resize(200, 100)

        self._button_confirm.clicked.connect(lambda checked=False: self.accept())
        self._button_cancel.clicked.connect(lambda checked=False: self.reject())
    
    def get_new_take(self):
        return self._widget_choose_take._radiobutton_new_take.isChecked()
    
    def show_dialog(self, state):
        self._widget_choose_take.update(state)
        return self.exec()


class SendShotsForValidationWidget(CustomPageWidget):

    def build(self):
        self._shot_list = ShotToSendList(self, self.session)
        self._combobox_task = QtWidgets.QComboBox()
        self._combobox_task.addItems(self.get_task_names())
        self._combobox_task.setCurrentText(self.get_target_task())
        refresh_icon = QtGui.QIcon(resources.get_icon(('icons.libreflow', 'refresh')))
        self._button_send = QtWidgets.QPushButton('Send')
        self._button_send.setFixedSize(50, 30)
        self._button_close = QtWidgets.QPushButton('Close')
        self._button_close.setFixedSize(50, 30)
        self._label_shot_count = QtWidgets.QLabel(f'{self._shot_list.topLevelItemCount() - 1} shot(s)')

        self.dialog_choose_take = ChooseTakeDialog(self)

        glo = QtWidgets.QGridLayout()
        glo.addWidget(self._combobox_task, 0, 0, 1, 5)
        glo.addWidget(self._shot_list, 1, 0, 1, 5)
        glo.addWidget(self._button_close, 2, 0)
        glo.addWidget(self._label_shot_count, 2, 2, QtCore.Qt.AlignCenter)
        glo.addWidget(self._button_send, 2, 4)
        glo.setContentsMargins(0, 0, 0, 0)
        glo.setSpacing(0)
        self.setLayout(glo)
    
        # Install callbacks
        self._button_send.clicked.connect(self._on_button_send_clicked)
        self._button_close.clicked.connect(self._on_button_close_clicked)
        self._combobox_task.currentTextChanged.connect(self._on_combobox_task_text_changed)
    
    def get_task_names(self):
        return self.session.cmds.Flow.get_value_choices(self.oid+'/task')
    
    def get_target_task(self):
        return self.session.cmds.Flow.get_value(self.oid+'/task')

    def set_target_task(self, task_name):
        self.session.cmds.Flow.set_value(self.oid+'/task', task_name)
    
    def update_list(self):
        self.session.cmds.Flow.call(self.oid+'/shots', 'touch', [], {})
        self._shot_list.update()
        shot_count = self._shot_list.topLevelItemCount() - 1
        self._label_shot_count.setText(f'{shot_count} shot(s)')

    def _on_button_send_clicked(self):
        self.session.cmds.Flow.run_action(self.oid, None)
        self.update_list()
    
    def _on_button_close_clicked(self):
        self.page.view.close()
    
    def _on_combobox_task_text_changed(self, text):
        self.set_target_task(text)
        self.update_list()



from libreflow.baseflow.runners import (
    CHOICES, CHOICES_ICONS,
    EditFileRunner,
    DefaultRunners as BaseDefaultRunners,
)

from ..resources import file_templates
from ..resources.icons import gui as _


CHOICES += ['plas', 'houdoo']


class Palas(EditFileRunner):

    @classmethod
    def supported_extensions(cls):
        return ['.plas']
    
    def show_terminal(self):
        return True
    
    def keep_terminal(self):
        return False


class Houdoo(EditFileRunner):

    @classmethod
    def supported_extensions(cls):
        return ['.houdoo']


class DefaultRunners(BaseDefaultRunners):

    def mapped_names(self, page_num=0, page_size=None):
        return CHOICES



from kabaret.app.ui.gui.widgets.flow.flow_view import QtCore, QtGui, QtWidgets, CustomPageWidget
from kabaret.app import resources

from .shot_list import ShotList

STYLESHEET = '''
    QPushButton:disabled {
        background-color: rgba(255, 255, 255, 0);
        color: rgba(255, 255, 255, 50);
    }'''


class DragDropWidget(QtWidgets.QWidget):

    def __init__(self, custom_widget):
        super(DragDropWidget, self).__init__(custom_widget)
        self.custom_widget = custom_widget

        self.setAcceptDrops(True)
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)

        self.frame = QtWidgets.QFrame()
        self.frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame.setStyleSheet('''
            background-color: #2b2b2b;
            border: 1px solid #22222b;
        ''')

        self.asset = QtWidgets.QWidget()
        asset_lo = QtWidgets.QVBoxLayout()
        icon = QtGui.QIcon(resources.get_icon(('icons.gui', 'file')))
        pixmap = icon.pixmap(QtCore.QSize(128, 128))
        self.icon_lbl = QtWidgets.QLabel('')
        self.icon_lbl.setPixmap(pixmap)
        self.label = QtWidgets.QLabel('Drop your files here')

        asset_lo.addWidget(self.icon_lbl, 0, QtCore.Qt.AlignCenter)
        asset_lo.addWidget(self.label, 1, QtCore.Qt.AlignCenter)
        self.asset.setLayout(asset_lo)
        
        glo = QtWidgets.QGridLayout()
        glo.setContentsMargins(0,0,0,0)
        glo.addWidget(self.frame, 0, 0, 3, 0)
        glo.addWidget(self.asset, 1, 0, QtCore.Qt.AlignCenter)
        self.setLayout(glo)

    def sizeHint(self):
        return QtCore.QSize(800, 493)

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls:
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        if event.mimeData().hasUrls:
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        if event.mimeData().hasUrls:
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
            links = []
            for url in event.mimeData().urls():
                links.append(str(url.toLocalFile()))
            self.custom_widget.add_files(links)
        else:
            event.ignore()


class ImportFilesWidget(CustomPageWidget):

    def build(self):
        self.setStyleSheet(STYLESHEET)

        self.reset_list()

        self.shot_list = ShotList(self)
        self.shots_count = QtWidgets.QLabel(self.shot_list.get_shots_count())
        self.dragdrop = DragDropWidget(self)
        
        self.button_settings = QtWidgets.QPushButton('Settings')
        if self.check_user_status() != 'Admin':
            self.button_settings.setEnabled(False)

        self.button_browse = QtWidgets.QPushButton('Browse')
        self.checkbox_selectall = QtWidgets.QCheckBox('Select all')
        self.checkbox_selectall.setEnabled(False)
        self.import_feedback = QtWidgets.QLabel('')
        self.button_import = QtWidgets.QPushButton('Import')

        if self.shot_list.topLevelItemCount() == 0:
            self.shot_list.hide()
            self.button_import.setEnabled(False)
        else:
            self.dragdrop.hide()
        
        glo = QtWidgets.QGridLayout()
        glo.addWidget(self.shots_count, 0, 0, 1, 6)
        glo.addWidget(self.shot_list, 1, 0, 1, 6)
        glo.addWidget(self.dragdrop, 1, 0, 1, 6)
        glo.addWidget(self.button_settings, 2, 0)
        glo.addWidget(self.button_browse, 2, 1)
        glo.addWidget(self.checkbox_selectall, 2, 2)
        glo.addWidget(self.import_feedback, 2, 4, QtCore.Qt.AlignRight)
        glo.addWidget(self.button_import, 2, 5)
        glo.setColumnStretch(2, 3)
        self.setLayout(glo)
    
        # Install callbacks
        # self.button_settings.clicked.connect(self._on_button_settings_clicked)
        self.button_browse.clicked.connect(self._on_button_browse_clicked)
        self.checkbox_selectall.stateChanged.connect(self._on_checkbox_selectall_state_changed)
        self.button_import.clicked.connect(self._on_button_import_clicked)

    def reset_list(self):
        return self.session.cmds.Flow.call(
            self.oid, 'clear_list', {}, {}
        )

    def refresh_shots_count(self):
        return self.shots_count.setText(self.shot_list.get_shots_count())

    def check_user_status(self):
        return self.session.cmds.Flow.call(
            self.oid, 'get_user_status', {}, {}
        )

    def get_shots(self):
        return self.session.cmds.Flow.get_mapped_oids(self.oid + '/shots_map')
    
    def get_shot_files(self, shot):
        return self.session.cmds.Flow.get_mapped_oids(shot + '/files_map')

    def remove_shot(self, name):
        return self.session.cmds.Flow.call(
            self.oid + '/shots_map', 'remove', [name], {}
        )

    def add_files(self, paths):
        # Error list:
        # file = unknown file
        # kitsu = shot not found on kitsu
        # init = shot needs to be init
        # rtk = waiting a retake file and not a base file
        # pack_only = single pack folder
        # dir_name = wrong dir name
        # dir_empty = dir is empty
        # no_track = only one version of the file can be created
        # overwrite = file will be overwrite
        # not_match = plas file_name don't have the same as dir
        # no_valid_files = after listdir, no valid files has been found
        # not_all_files = can't create because not all files has been found

        self.button_import.setEnabled(False)
        self.import_feedback.setText('Waiting...')
        QtWidgets.QApplication.processEvents()
        QtWidgets.QApplication.processEvents()

        check_files = self.session.cmds.Flow.call(
            self.oid, 'check_files', [paths], {}
        )

        if check_files != {'files': [], 'shots': []}:
            # Dialog when incorrect files
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Warning)

            body = '''<b>Warning</b><br><br>
            Some files/shots can\'t be imported.<br><br>
            '''

            for f in check_files['files']:
                display_error = ''
                if f['error'] == 'file':
                    display_error = 'Unknown file'
                if f['error'] == 'kitsu':
                    display_error = 'Shot not found in Kitsu'
                if f['error'] == 'rtk':
                    display_error = 'Shot needs a retake file, not a base file'
                if f['error'] == 'init':
                    display_error = 'Shot needs to be initialized'
                if f['error'] == 'pack_only':
                    display_error = 'You should send the shot folder'
                if f['error'] == 'dir_name':
                    display_error = 'No target for this directory name'
                body = body + '''
                - {name} : {error}<br>
                '''.format(name=f['name'], error=display_error)
            
            body = body + '''
            <br>
            '''

            for s in check_files['shots']:
                if s['error'] == '':
                    body = body + '''
                    - {name}<br>
                    '''.format(name=s['shot'])
                else:
                    shot_display_error = ''
                    if s['error'] == 'kitsu':
                        shot_display_error = 'Not found in Kitsu'
                    if s['error'] == 'init':
                        shot_display_error = 'Shot needs to be initialized'
                    if s['error'] == 'dir_empty':
                        shot_display_error = 'Folder is empty'
                    if s['error'] == 'no_valid_files':
                        shot_display_error = 'Has no valid files'
                    if s['error'] == 'not_all_files':
                        shot_display_error = 'Missing necessary files'
                    body = body + '''
                    - {name} : {error}<br>
                    '''.format(name=s['shot'], error=shot_display_error)

                for f in s['files']:
                    display_error = ''
                    if f['error'] == 'file':
                        display_error = 'Unknown file'
                    if f['error'] == 'no_track':
                        display_error = 'Only one version of this file can be created'
                    if f['error'] == 'overwrite':
                        display_error = 'This file will be overwritten if imported'
                        self.shot_list.overwrite.append(s['shot'])
                    if f['error'] == 'not_match':
                        display_error = 'File name is not equal as folder'
                    if f['error'] == 'rtk':
                        display_error = 'Retake file is needed, not a base file'
                    if f['error'] == 'init':
                        display_error = 'Base file is needed, not a retake'
                    body = body + '''
                    &nbsp;&nbsp;&nbsp;&nbsp;- {name} : {error}<br>
                    '''.format(name=f['name'], error=display_error)

            msg.setText(body)

            msg.setWindowTitle('libreflow')
            msg.setWindowIcon(
                resources.get_icon(('icons.gui', 'kabaret_icon'))
            )
            msg.exec_()

        if self.get_shots() != []:
            if self.shot_list.isVisible() == False:
                self.shot_list.show()
                self.dragdrop.hide()
                self.checkbox_selectall.setCheckState(QtCore.Qt.Checked)
                self.button_import.setEnabled(True)
        
            self.shot_list.refresh()
        
        self.import_feedback.setText('')

    def _on_button_browse_clicked(self):
        dialog = QtWidgets.QFileDialog(self)
        dialog.setStyleSheet('''
            QLineEdit:focus {
                border: none;
                padding: 0px;
                padding-left: 2px;
            }
        ''')
        dialog.setFileMode(QtWidgets.QFileDialog.Directory)
        dialog.setOption(QtWidgets.QFileDialog.DontUseNativeDialog, True)

        file_view = dialog.findChild(QtWidgets.QListView, 'listView')
        if file_view:
            file_view.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
            file_view.setSelectionRectVisible(True)
        f_tree_view = dialog.findChild(QtWidgets.QTreeView)
        if f_tree_view:
            f_tree_view.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)

        if dialog.exec():
            paths = dialog.selectedFiles()
            self.add_files(paths)

    def _on_checkbox_selectall_state_changed(self, state):
        for i in range(self.shot_list.topLevelItemCount()):
            state = QtCore.Qt.CheckState(state)
            self.shot_list.topLevelItem(i).setCheckState(0, state)

    def _on_button_import_clicked(self):
        shots = []
        for i in range(self.shot_list.topLevelItemCount()):
            shot_item = self.shot_list.topLevelItem(i)

            if shot_item.checkState(0) == QtCore.Qt.Checked:
                shots.append(shot_item.text(0))
        
        if self.shot_list.overwrite != []:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setText('''<b>Warning</b><br>Some files will be overwritten''')
            msg.setInformativeText('Are you sure to import?')
            msg.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
            msg.setWindowTitle('libreflow')
            msg.setWindowIcon(
                resources.get_icon(('icons.gui', 'kabaret_icon'))
            )
            ret = msg.exec_()
            
            if ret == QtWidgets.QMessageBox.No:
                return

        self.button_import.setEnabled(False)
        self.import_feedback.setText('Importing...')
        QtWidgets.QApplication.processEvents()
        QtWidgets.QApplication.processEvents()
        
        self.session.cmds.Flow.call(
            self.oid, 'import_shot_files', [shots], {}
        )

        self.shot_list.refresh()

        self.import_feedback.setText('Completed')


from kabaret import flow
from libreflow.baseflow.task import ManagedTask, ManagedTaskCollection


class Task(ManagedTask):
    
    @classmethod
    def get_source_display(cls, oid):
        split = oid.split('/')
        return f'{split[3]} · {split[5]} · {split[7]}'


class Tasks(ManagedTaskCollection):

    def mapped_items(self, page_num=0, page_size=None):
        items = super(Tasks, self).mapped_items(
            page_num, page_size
        )

        mgr = self.root().project().get_task_manager()
        current_site = self.root().project().get_current_site()
        visible_tasks = current_site.visible_tasks.get()

        items = [
            i for i in items
            if not mgr.default_tasks[i.name()].assignation_enabled.get()
            or i.name() in visible_tasks
        ]
        return items
